// Ansteuerung eines HD44780 kompatiblen LCD im 4-Bit-Interfacemodus
// http://www.mikrocontroller.net/articles/HD44780
// http://www.mikrocontroller.net/articles/AVR-GCC-Tutorial/LCD-Ansteuerung
//
// Modifizierte Version von Matthias Rock
 
#include <avr/io.h>
#include "lcd-routines.h"
#include <util/delay.h>
#include <string.h>
 
////////////////////////////////////////////////////////////////////////////////
// Erzeugt einen Enable-Puls
static void lcd_enable(void) {

    LCD_PORT |= (1<<LCD_EN);     			// Enable auf 1 setzen
    _delay_us(LCD_ENABLE_US);  				// kurze Pause
    LCD_PORT &= ~(1<<LCD_EN);    			// Enable auf 0 setzen

}
 
////////////////////////////////////////////////////////////////////////////////
// Sendet eine 4-bit Ausgabeoperation an das LCD
static void lcd_out(uint8_t data) {

    data &= 0xF0;                       	// obere 4 Bit maskieren
    LCD_PORT &= ~(0xF0>>(4-LCD_DB));    	// Maske lschen
    LCD_PORT |= (data>>(4-LCD_DB));     	// Bits setzen
    lcd_enable();

}
 
////////////////////////////////////////////////////////////////////////////////
// Initialisierung: muss ganz am Anfang des Programms aufgerufen werden.
void lcd_init(void) {

    // verwendete Pins auf Ausgang schalten
    uint8_t pins = (0x0F << LCD_DB) |    	// 4 Datenleitungen
                   (1<<LCD_RS) |          	// R/S Leitung
                   (1<<LCD_EN);           	// Enable Leitung
    LCD_DDR |= pins;
 
    // initial alle Ausgnge auf Null
    LCD_PORT &= ~pins;
 
    // warten auf die Bereitschaft des LCD
    _delay_ms( LCD_BOOTUP_MS );
 
    // Soft-Reset muss 3mal hintereinander gesendet werden zur Initialisierung
    lcd_out( LCD_SOFT_RESET );
    _delay_ms( LCD_SOFT_RESET_MS1 );
 
    lcd_enable();
    _delay_ms( LCD_SOFT_RESET_MS2 );
 
    lcd_enable();
    _delay_ms( LCD_SOFT_RESET_MS3 );
 
    // 4-bit Modus aktivieren 
    lcd_out( LCD_SET_FUNCTION |
             LCD_FUNCTION_4BIT );
    _delay_ms( LCD_SET_4BITMODE_MS );
 
    // 4-bit Modus / 2 Zeilen / 5x7
    lcd_command( LCD_SET_FUNCTION |
                 LCD_FUNCTION_4BIT |
                 LCD_FUNCTION_2LINE |
                 LCD_FUNCTION_5X7 );
 
    // Display ein / Cursor aus / Blinken aus
    lcd_command( LCD_SET_DISPLAY |
                 LCD_DISPLAY_ON |
                 LCD_CURSOR_OFF |
                 LCD_BLINKING_OFF); 
 
    // Cursor inkrement / kein Scrollen
    lcd_command( LCD_SET_ENTRY |
                 LCD_ENTRY_INCREASE |
                 LCD_ENTRY_NOSHIFT );
 
    lcd_clear();

}
 
////////////////////////////////////////////////////////////////////////////////
// Sendet ein Datenbyte an das LCD
void lcd_data(uint8_t data) {

	LCD_PORT |= (1<<LCD_RS);    		// RS auf 1 setzen
    lcd_out( data );            		// zuerst die oberen, 
    lcd_out( data<<4 );         		// dann die unteren 4 Bit senden
    _delay_us(LCD_WRITEDATA_US);

}
 
////////////////////////////////////////////////////////////////////////////////
// Sendet einen Befehl an das LCD
void lcd_command(uint8_t data) {

    LCD_PORT &= ~(1<<LCD_RS);    		// RS auf 0 setzen
    lcd_out( data );             		// zuerst die oberen, 
    lcd_out( data<<4 );           		// dann die unteren 4 Bit senden
    _delay_us( LCD_COMMAND_US );

}
 
////////////////////////////////////////////////////////////////////////////////
// Sendet den Befehl zur Lschung des Displays
void lcd_clear( void ) {

    lcd_command(LCD_CLEAR_DISPLAY);
    _delay_ms(LCD_CLEAR_DISPLAY_MS);

}
 
////////////////////////////////////////////////////////////////////////////////
// Sendet den Befehl: Cursor Home
void lcd_home( void ) {

    lcd_command(LCD_CURSOR_HOME);
    _delay_ms(LCD_CURSOR_HOME_MS);

}
 
////////////////////////////////////////////////////////////////////////////////
// Setzt den Cursor in die Zeile 1...4 und Spalte 1...20 
uint8_t lcd_setcursor(uint8_t zeile, uint8_t spalte) {

    uint8_t data;
	
	// Wenn Zeile oder Spalte auerhalb des erlaubten Bereichs
	if(zeile < 1 || zeile > 4 || spalte < 1 || spalte > 20) return 0;

	switch(zeile) {

        // 1. Zeile
		case 1:
            data = LCD_SET_DDADR + LCD_DDADR_LINE1 + spalte - 1;
            break;
        // 2. Zeile
		case 2:
            data = LCD_SET_DDADR + LCD_DDADR_LINE2 + spalte - 1;
            break;
        // 3. Zeile
		case 3:
            data = LCD_SET_DDADR + LCD_DDADR_LINE3 + spalte - 1;
            break;
        // 4. Zeile
		case 4:
            data = LCD_SET_DDADR + LCD_DDADR_LINE4 + spalte - 1;
            break;
        default:
            return 0;

    }
    lcd_command(data);

	return 1;

}
 
////////////////////////////////////////////////////////////////////////////////
// Schreibt einen String auf das LCD in die Zeile 1...4 (ab Spalte 1...20)
void lcd_string(uint8_t zeile, uint8_t spalte, const char *data) {

	uint8_t string_length, iterations; 
	   
	// Cursor an die entsprechende Stelle setzen, falls Position gltig, sonst abbrechen
	if(!lcd_setcursor(zeile, spalte)) return;			
	
	string_length = strlen(data);						// Lnge des Strings bestimmen

	// Wenn der String nicht komplett in die Zeile passt
	if(spalte + string_length > 21) {
	
		iterations = 21 - spalte;

	}
	// Wenn der String komplett in die Zeile passt
	else {
	
		iterations = string_length;
	
	}
	while(*data != '\0' && iterations > 0) {
	
		lcd_data(*data);
		data++;
		iterations--;

	}

}
 
////////////////////////////////////////////////////////////////////////////////
// Schreibt ein Zeichen in den Character Generator RAM
 
void lcd_generatechar(uint8_t code, const uint8_t *data) {

    // Startposition des Zeichens einstellen
    lcd_command(LCD_SET_CGADR | (code<<3));
 
    // Bitmuster bertragen
    for (uint8_t i = 0; i < 8; i++) lcd_data(data[i]);

}
